@@ -27,15 +27,18 @@
 
 #include <pyglib.h>
 #include "pygobject-private.h"
+#include "pygflags.h"
 
-#define GET_INT_VALUE(x) (((PyIntObject*)x)->ob_ival)
+GQuark pygflags_class_key;
+
+PYGLIB_DEFINE_TYPE("gobject.GFlags", PyGFlags_Type, PyGFlags);
 
 static PyObject *
 pyg_flags_richcompare(PyGFlags *self, PyObject *other, int op)
 {
     static char warning[256];
 
-    if (!PyInt_Check(other)) {
+    if (!_PyLong_Check(other)) {
         Py_INCREF(Py_NotImplemented);
         return Py_NotImplemented;
     }
@@ -89,17 +92,17 @@ pyg_flags_repr(PyGFlags *self)
     char *tmp, *retval;
     PyObject *pyretval;
   
-    tmp = generate_repr(self->gtype, self->parent.ob_ival);
+    tmp = generate_repr(self->gtype, _PyLong_AS_LONG(self));
 
     if (tmp)
         retval = g_strdup_printf("<flags %s of type %s>", tmp,
                                  g_type_name(self->gtype));
     else
-        retval = g_strdup_printf("<flags %ld of type %s>", self->parent.ob_ival,
+        retval = g_strdup_printf("<flags %ld of type %s>", _PyLong_AS_LONG(self),
                                  g_type_name(self->gtype));
     g_free(tmp);
     
-    pyretval = PyString_FromString(retval);
+    pyretval = _PyUnicode_FromString(retval);
     g_free(retval);
   
     return pyretval;
@@ -148,7 +151,7 @@ pyg_flags_new(PyTypeObject *type, PyObject *args, PyObject *kwargs)
 
     g_type_class_unref(eclass);
     
-    pyint = PyInt_FromLong(value);
+    pyint = _PyLong_FromLong(value);
     ret = PyDict_GetItem(values, pyint);
     Py_DECREF(pyint);
     Py_DECREF(values);
@@ -171,13 +174,13 @@ pyg_flags_from_gtype (GType gtype, int value)
     if (pyclass == NULL) {
 	pyclass = pyg_flags_add(NULL, g_type_name(gtype), NULL, gtype);
 	if (!pyclass)
-	    return PyInt_FromLong(value);
+	    return _PyLong_FromLong(value);
     }
 
     
     values = PyDict_GetItemString(((PyTypeObject *)pyclass)->tp_dict,
 				  "__flags_values__");
-    pyint = PyInt_FromLong(value);
+    pyint = _PyLong_FromLong(value);
     retval = PyDict_GetItem(values, pyint);
     Py_DECREF(pyint);
 
@@ -187,7 +190,11 @@ pyg_flags_from_gtype (GType gtype, int value)
 	retval = ((PyTypeObject *)pyclass)->tp_alloc((PyTypeObject *)pyclass, 0);
 	g_assert(retval != NULL);
 	
-	((PyIntObject*)retval)->ob_ival = value;
+#if PY_VERSION_HEX >= 0x03000000
+#   warning "FIXME: figure out how to subclass long"    
+#else
+	((_PyLongObject*)retval)->ob_ival = value;
+#endif    
 	((PyGFlags*)retval)->gtype = gtype;
     } else {
 	Py_INCREF(retval);
@@ -232,7 +239,7 @@ pyg_flags_add (PyObject *   module,
     if (module) {
         PyDict_SetItemString(((PyTypeObject *)stub)->tp_dict,
                              "__module__",
-                             PyString_FromString(PyModule_GetName(module)));
+                             _PyUnicode_FromString(PyModule_GetName(module)));
     
           /* Add it to the module name space */
         PyModule_AddObject(module, (char*)typename, stub);
@@ -252,10 +259,14 @@ pyg_flags_add (PyObject *   module,
       PyObject *item, *intval;
       
       item = ((PyTypeObject *)stub)->tp_alloc((PyTypeObject *)stub, 0);
-      ((PyIntObject*)item)->ob_ival = eclass->values[i].value;
+#if PY_VERSION_HEX >= 0x03000000
+#   warning "FIXME: figure out how to subclass long"    
+#else
+      ((_PyLongObject*)item)->ob_ival = eclass->values[i].value;
+#endif    
       ((PyGFlags*)item)->gtype = gtype;
             
-      intval = PyInt_FromLong(eclass->values[i].value);
+      intval = _PyLong_FromLong(eclass->values[i].value);
       PyDict_SetItem(values, intval, item);
       Py_DECREF(intval);
 
@@ -285,32 +296,32 @@ static PyObject *
 pyg_flags_and(PyGFlags *a, PyGFlags *b)
 {
 	if (!PyGFlags_Check(a) || !PyGFlags_Check(b))
-		return PyInt_Type.tp_as_number->nb_and((PyObject*)a,
+		return _PyLong_Type.tp_as_number->nb_and((PyObject*)a,
 						       (PyObject*)b);
 	
 	return pyg_flags_from_gtype(a->gtype,
-				    GET_INT_VALUE(a) & GET_INT_VALUE(b));
+				    _PyLong_AS_LONG(a) & _PyLong_AS_LONG(b));
 }
 
 static PyObject *
 pyg_flags_or(PyGFlags *a, PyGFlags *b)
 {
 	if (!PyGFlags_Check(a) || !PyGFlags_Check(b))
-		return PyInt_Type.tp_as_number->nb_or((PyObject*)a,
+		return _PyLong_Type.tp_as_number->nb_or((PyObject*)a,
 						      (PyObject*)b);
 
-	return pyg_flags_from_gtype(a->gtype, GET_INT_VALUE(a) | GET_INT_VALUE(b));
+	return pyg_flags_from_gtype(a->gtype, _PyLong_AS_LONG(a) | _PyLong_AS_LONG(b));
 }
 
 static PyObject *
 pyg_flags_xor(PyGFlags *a, PyGFlags *b)
 {
 	if (!PyGFlags_Check(a) || !PyGFlags_Check(b))
-		return PyInt_Type.tp_as_number->nb_xor((PyObject*)a,
+		return _PyLong_Type.tp_as_number->nb_xor((PyObject*)a,
 						       (PyObject*)b);
 
 	return pyg_flags_from_gtype(a->gtype,
-				    GET_INT_VALUE(a) ^ GET_INT_VALUE(b));
+				    _PyLong_AS_LONG(a) ^ _PyLong_AS_LONG(b));
 
 }
 
@@ -333,9 +344,9 @@ pyg_flags_get_first_value_name(PyGFlags *self, void *closure)
   
   flags_class = g_type_class_ref(self->gtype);
   g_assert(G_IS_FLAGS_CLASS(flags_class));
-  flags_value = g_flags_get_first_value(flags_class, self->parent.ob_ival);
+  flags_value = g_flags_get_first_value(flags_class, _PyLong_AS_LONG(self));
   if (flags_value)
-      retval = PyString_FromString(flags_value->value_name);
+      retval = _PyUnicode_FromString(flags_value->value_name);
   else {
       retval = Py_None;
       Py_INCREF(Py_None);
@@ -355,9 +366,9 @@ pyg_flags_get_first_value_nick(PyGFlags *self, void *closure)
   flags_class = g_type_class_ref(self->gtype);
   g_assert(G_IS_FLAGS_CLASS(flags_class));
 
-  flags_value = g_flags_get_first_value(flags_class, self->parent.ob_ival);
+  flags_value = g_flags_get_first_value(flags_class, _PyLong_AS_LONG(self));
   if (flags_value)
-      retval = PyString_FromString(flags_value->value_nick);
+      retval = _PyUnicode_FromString(flags_value->value_nick);
   else {
       retval = Py_None;
       Py_INCREF(Py_None);
@@ -379,8 +390,8 @@ pyg_flags_get_value_names(PyGFlags *self, void *closure)
   
   retval = PyList_New(0);
   for (i = 0; i < flags_class->n_values; i++)
-      if ((self->parent.ob_ival & flags_class->values[i].value) == flags_class->values[i].value)
-	  PyList_Append(retval, PyString_FromString(flags_class->values[i].value_name));
+      if ((_PyLong_AS_LONG(self) & flags_class->values[i].value) == flags_class->values[i].value)
+	  PyList_Append(retval, _PyUnicode_FromString(flags_class->values[i].value_name));
 
   g_type_class_unref(flags_class);
 
@@ -399,8 +410,8 @@ pyg_flags_get_value_nicks(PyGFlags *self, void *closure)
   
   retval = PyList_New(0);
   for (i = 0; i < flags_class->n_values; i++)
-      if ((self->parent.ob_ival & flags_class->values[i].value) == flags_class->values[i].value)
-	  PyList_Append(retval, PyString_FromString(flags_class->values[i].value_nick));
+      if ((_PyLong_AS_LONG(self) & flags_class->values[i].value) == flags_class->values[i].value)
+	  PyList_Append(retval, _PyUnicode_FromString(flags_class->values[i].value_nick));
 
   g_type_class_unref(flags_class);
 
@@ -456,44 +467,18 @@ static PyNumberMethods pyg_flags_as_number = {
 	0,					/* nb_inplace_true_divide */
 };
 
-PyTypeObject PyGFlags_Type = {
-	PyObject_HEAD_INIT(NULL)
-	0,
-	"gobject.GFlags",
-	sizeof(PyGFlags),
-	0,
-	0,					  /* tp_dealloc */
-	0,                			  /* tp_print */
-	0,					  /* tp_getattr */
-	0,					  /* tp_setattr */
-	0,					  /* tp_compare */
-	(reprfunc)pyg_flags_repr,		  /* tp_repr */
-	&pyg_flags_as_number,			  /* tp_as_number */
-	0,					  /* tp_as_sequence */
-	0,					  /* tp_as_mapping */
-	0,					  /* tp_hash */
-        0,					  /* tp_call */
-	(reprfunc)pyg_flags_repr,		  /* tp_str */
-	0,					  /* tp_getattro */
-	0,					  /* tp_setattro */
-	0,				  	  /* tp_as_buffer */
-	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /* tp_flags */
-	0,      				  /* tp_doc */
-	0,					  /* tp_traverse */
-	0,					  /* tp_clear */
-	(richcmpfunc)pyg_flags_richcompare,	  /* tp_richcompare */
-	0,					  /* tp_weaklistoffset */
-	0,					  /* tp_iter */
-	0,					  /* tp_iternext */
-	0,					  /* tp_methods */
-	0,					  /* tp_members */
-	pyg_flags_getsets,			  /* tp_getset */
-	0,	  				  /* tp_base */
-	0,					  /* tp_dict */
-	0,					  /* tp_descr_get */
-	0,					  /* tp_descr_set */
-	0,					  /* tp_dictoffset */
-	0,					  /* tp_init */
-	0,					  /* tp_alloc */
-	pyg_flags_new,				  /* tp_new */
-};
+void
+pygobject_flags_register_types(PyObject *d)
+{
+    pygflags_class_key = g_quark_from_static_string("PyGFlags::class");
+
+    PyGFlags_Type.tp_base = &_PyLong_Type;
+    PyGFlags_Type.tp_repr = (reprfunc)pyg_flags_repr;
+    PyGFlags_Type.tp_as_number = &pyg_flags_as_number;
+    PyGFlags_Type.tp_str = (reprfunc)pyg_flags_repr;
+    PyGFlags_Type.tp_flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE;
+    PyGFlags_Type.tp_richcompare = (richcmpfunc)pyg_flags_richcompare;
+    PyGFlags_Type.tp_getset = pyg_flags_getsets;
+    PyGFlags_Type.tp_new = pyg_flags_new;
+    PYGOBJECT_REGISTER_GTYPE(d, PyGFlags_Type, "GFlags", G_TYPE_FLAGS);
+}
