@@ -26,11 +26,16 @@
 
 #include <pyglib.h>
 #include "pygobject-private.h"
+#include "pygpointer.h"
+
+GQuark pygpointer_class_key;
+
+PYGLIB_DEFINE_TYPE("gobject.GPointer", PyGPointer_Type, PyGPointer);
 
 static void
 pyg_pointer_dealloc(PyGPointer *self)
 {
-    self->ob_type->tp_free((PyObject *)self);
+    Py_TYPE(self)->tp_free((PyObject *)self);
 }
 
 static int
@@ -54,7 +59,7 @@ pyg_pointer_repr(PyGPointer *self)
 
     g_snprintf(buf, sizeof(buf), "<%s at 0x%lx>", g_type_name(self->gtype),
 	       (long)self->pointer);
-    return PyString_FromString(buf);
+    return _PyUnicode_FromString(buf);
 }
 
 static int
@@ -68,7 +73,8 @@ pyg_pointer_init(PyGPointer *self, PyObject *args, PyObject *kwargs)
     self->pointer = NULL;
     self->gtype = 0;
 
-    g_snprintf(buf, sizeof(buf), "%s can not be constructed", self->ob_type->tp_name);
+    g_snprintf(buf, sizeof(buf), "%s can not be constructed",
+	       Py_TYPE(self)->tp_name);
     PyErr_SetString(PyExc_NotImplementedError, buf);
     return -1;
 }
@@ -79,52 +85,6 @@ pyg_pointer_free(PyObject *op)
   PyObject_FREE(op);
 }
 
-PyTypeObject PyGPointer_Type = {
-    PyObject_HEAD_INIT(NULL)
-    0,                                  /* ob_size */
-    "gobject.GPointer",                 /* tp_name */
-    sizeof(PyGPointer),                 /* tp_basicsize */
-    0,                                  /* tp_itemsize */
-    /* methods */
-    (destructor)pyg_pointer_dealloc,    /* tp_dealloc */
-    (printfunc)0,                       /* tp_print */
-    (getattrfunc)0,                     /* tp_getattr */
-    (setattrfunc)0,                     /* tp_setattr */
-    (cmpfunc)pyg_pointer_compare,       /* tp_compare */
-    (reprfunc)pyg_pointer_repr,         /* tp_repr */
-    0,                                  /* tp_as_number */
-    0,                                  /* tp_as_sequence */
-    0,                                  /* tp_as_mapping */
-    (hashfunc)pyg_pointer_hash,         /* tp_hash */
-    (ternaryfunc)0,                     /* tp_call */
-    (reprfunc)0,                        /* tp_str */
-    (getattrofunc)0,			/* tp_getattro */
-    (setattrofunc)0,                    /* tp_setattro */
-    0,					/* tp_as_buffer */
-    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,	/* tp_flags */
-    NULL, /* Documentation string */
-    (traverseproc)0,			/* tp_traverse */
-    (inquiry)0,				/* tp_clear */
-    (richcmpfunc)0,			/* tp_richcompare */
-    0,					/* tp_weaklistoffset */
-    (getiterfunc)0,			/* tp_iter */
-    (iternextfunc)0,			/* tp_iternext */
-    0,					/* tp_methods */
-    0,					/* tp_members */
-    0,					/* tp_getset */
-    (PyTypeObject *)0,			/* tp_base */
-    (PyObject *)0,			/* tp_dict */
-    0,					/* tp_descr_get */
-    0,					/* tp_descr_set */
-    0,					/* tp_dictoffset */
-    (initproc)pyg_pointer_init,		/* tp_init */
-    (allocfunc)0,			/* tp_alloc */
-    (newfunc)0,				/* tp_new */
-    (freefunc)pyg_pointer_free,		/* tp_free */
-    (inquiry)0,				/* tp_is_gc */
-    (PyObject *)0,			/* tp_bases */
-};
-
 /**
  * pyg_register_pointer:
  * @dict: the module dictionary to store the wrapper class.
@@ -148,7 +108,7 @@ pyg_register_pointer(PyObject *dict, const gchar *class_name,
 
     if (!type->tp_dealloc) type->tp_dealloc = (destructor)pyg_pointer_dealloc;
 
-    type->ob_type = &PyType_Type;
+    Py_TYPE(type) = &PyType_Type;
     type->tp_base = &PyGPointer_Type;
 
     if (PyType_Ready(type) < 0) {
@@ -209,3 +169,18 @@ pyg_pointer_new(GType pointer_type, gpointer pointer)
 
     return (PyObject *)self;
 }
+
+void
+pygobject_pointer_register_types(PyObject *d)
+{
+    pygpointer_class_key     = g_quark_from_static_string("PyGPointer::class");
+
+    PyGPointer_Type.tp_dealloc = (destructor)pyg_pointer_dealloc;
+    PyGPointer_Type.tp_compare = (cmpfunc)pyg_pointer_compare;
+    PyGPointer_Type.tp_repr = (reprfunc)pyg_pointer_repr;
+    PyGPointer_Type.tp_hash = (hashfunc)pyg_pointer_hash;
+    PyGPointer_Type.tp_flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE;
+    PyGPointer_Type.tp_init = (initproc)pyg_pointer_init;
+    PyGPointer_Type.tp_free = (freefunc)pyg_pointer_free;
+    PYGOBJECT_REGISTER_GTYPE(d, PyGPointer_Type, "GPointer", G_TYPE_POINTER); 
+}
