@@ -20,9 +20,15 @@
  * USA
  */
 
+#ifdef HAVE_CONFIG_H
+#  include <config.h>
+#endif
+
 #include <pyglib.h>
 
 #include "pygobject-private.h"
+#include "pygparamspec.h"
+#include "pygtype.h"
 
 /* -------------- __gtype__ objects ---------------------------- */
 
@@ -31,6 +37,8 @@ typedef struct {
     GType type;
 } PyGTypeWrapper;
 
+PYGLIB_DEFINE_TYPE("gobject.GType", PyGTypeWrapper_Type, PyGTypeWrapper);
+
 static int
 pyg_type_wrapper_compare(PyGTypeWrapper *self, PyGTypeWrapper *v)
 {
@@ -53,7 +61,7 @@ pyg_type_wrapper_repr(PyGTypeWrapper *self)
 
     g_snprintf(buf, sizeof(buf), "<GType %s (%lu)>",
 	       name?name:"invalid", (unsigned long int) self->type);
-    return PyString_FromString(buf);
+    return _PyUnicode_FromString(buf);
 }
 
 static void
@@ -79,7 +87,7 @@ static PyObject *
 _wrap_g_type_wrapper__get_name(PyGTypeWrapper *self, void *closure)
 {
    const char *name = g_type_name(self->type);
-   return PyString_FromString(g_strdup(name ? name : "invalid"));
+   return _PyUnicode_FromString(g_strdup(name ? name : "invalid"));
 }
 
 static PyObject *
@@ -131,7 +139,7 @@ _wrap_g_type_wrapper__get_interfaces(PyGTypeWrapper *self, void *closure)
 static PyObject *
 _wrap_g_type_wrapper__get_depth(PyGTypeWrapper *self, void *closure)
 {
-  return PyInt_FromLong(g_type_depth(self->type));
+  return _PyLong_FromLong(g_type_depth(self->type));
 }
 
 static PyGetSetDef _PyGTypeWrapper_getsets[] = {
@@ -266,46 +274,6 @@ pyg_type_wrapper_init(PyGTypeWrapper *self, PyObject *args, PyObject *kwargs)
     return 0;
 }
 
-PyTypeObject PyGTypeWrapper_Type = {
-    PyObject_HEAD_INIT(NULL)
-    0,
-    "gobject.GType",
-    sizeof(PyGTypeWrapper),
-    0,
-    (destructor)pyg_type_wrapper_dealloc,
-    (printfunc)0,
-    (getattrfunc)0,
-    (setattrfunc)0,
-    (cmpfunc)pyg_type_wrapper_compare,
-    (reprfunc)pyg_type_wrapper_repr,
-    0,
-    0,
-    0,
-    (hashfunc)pyg_type_wrapper_hash,
-    (ternaryfunc)0,
-    (reprfunc)0,
-    (getattrofunc)0,		
-    (setattrofunc)0,		
-    0,				
-    Py_TPFLAGS_DEFAULT, 
-    NULL,
-    (traverseproc)0,		
-    (inquiry)0,
-    (richcmpfunc)0,	
-    0,             
-    (getiterfunc)0,
-    (iternextfunc)0,
-    _PyGTypeWrapper_methods,
-    0,
-    _PyGTypeWrapper_getsets,
-    NULL,		
-    NULL,		
-    (descrgetfunc)0,	
-    (descrsetfunc)0,	
-    0,                 
-    (initproc)pyg_type_wrapper_init
-};
-
 /**
  * pyg_type_wrapper_new:
  * type: a GType
@@ -355,7 +323,7 @@ pyg_type_from_object(PyObject *obj)
     if (PyType_Check(obj)) {
 	PyTypeObject *tp = (PyTypeObject *)obj;
 
-	if (tp == &PyInt_Type)
+	if (tp == &_PyLong_Type)
 	    return G_TYPE_INT;
 	else if (tp == &PyBool_Type)
 	    return G_TYPE_BOOLEAN;
@@ -363,19 +331,19 @@ pyg_type_from_object(PyObject *obj)
 	    return G_TYPE_LONG;
 	else if (tp == &PyFloat_Type)
 	    return G_TYPE_DOUBLE;
-	else if (tp == &PyString_Type)
+	else if (tp == &_PyUnicode_Type)
 	    return G_TYPE_STRING;
 	else if (tp == &PyBaseObject_Type)
 	    return PY_TYPE_OBJECT;
     }
 
-    if (obj->ob_type == &PyGTypeWrapper_Type) {
+    if (Py_TYPE(obj) == &PyGTypeWrapper_Type) {
 	return ((PyGTypeWrapper *)obj)->type;
     }
 
     /* handle strings */
-    if (PyString_Check(obj)) {
-	gchar *name = PyString_AsString(obj);
+    if (_PyUnicode_Check(obj)) {
+	gchar *name = _PyUnicode_AsString(obj);
 	
 	type = _pyg_type_from_name(name);
 	if (type != 0) {
@@ -387,7 +355,7 @@ pyg_type_from_object(PyObject *obj)
     gtype = PyObject_GetAttrString(obj, "__gtype__");
 
     if (gtype) {
-	if (gtype->ob_type == &PyGTypeWrapper_Type) {
+	if (Py_TYPE(gtype) == &PyGTypeWrapper_Type) {
 	    type = ((PyGTypeWrapper *)gtype)->type;
 	    Py_DECREF(gtype);
 	    return type;
@@ -425,12 +393,18 @@ pyg_enum_get_value(GType enum_type, PyObject *obj, gint *val)
     if (!obj) {
 	*val = 0;
 	res = 0;
-    } else if (PyInt_Check(obj)) {
-	*val = PyInt_AsLong(obj);
+    } else if (_PyLong_Check(obj)) {
+	*val = _PyLong_AsLong(obj);
 	res = 0;
-    } else if (PyString_Check(obj)) {
+
+	if (PyObject_TypeCheck(obj, &PyGEnum_Type) && ((PyGEnum *) obj)->gtype != enum_type) {
+	    g_warning("expected enumeration type %s, but got %s instead",
+		      g_type_name(enum_type),
+		      g_type_name(((PyGEnum *) obj)->gtype));
+	}
+    } else if (_PyUnicode_Check(obj)) {
 	GEnumValue *info;
-	char *str = PyString_AsString(obj);
+	char *str = _PyUnicode_AsString(obj);
 	
 	if (enum_type != G_TYPE_NONE)
 	    eclass = G_ENUM_CLASS(g_type_class_ref(enum_type));
@@ -482,15 +456,15 @@ pyg_flags_get_value(GType flag_type, PyObject *obj, gint *val)
     if (!obj) {
 	*val = 0;
 	res = 0;
-    } else if (PyInt_Check(obj)) {
-	*val = PyInt_AsLong(obj);
+    } else if (_PyLong_Check(obj)) {
+	*val = _PyLong_AsLong(obj);
 	res = 0;
     } else if (PyLong_Check(obj)) {
         *val = PyLong_AsLongLong(obj);
         res = 0;
-    } else if (PyString_Check(obj)) {
+    } else if (_PyUnicode_Check(obj)) {
 	GFlagsValue *info;
-	char *str = PyString_AsString(obj);
+	char *str = _PyUnicode_AsString(obj);
 
 	if (flag_type != G_TYPE_NONE)
 	    fclass = G_FLAGS_CLASS(g_type_class_ref(flag_type));
@@ -526,7 +500,7 @@ pyg_flags_get_value(GType flag_type, PyObject *obj, gint *val)
 
 	for (i = 0; i < len; i++) {
 	    PyObject *item = PyTuple_GetItem(obj, i);
-	    char *str = PyString_AsString(item);
+	    char *str = _PyUnicode_AsString(item);
 	    GFlagsValue *info = g_flags_get_value_by_name(fclass, str);
 
 	    if (!info)
@@ -633,7 +607,7 @@ pyg_value_array_from_pyobject(GValue *value,
 	else if (item == Py_None)
 	    type = G_TYPE_POINTER; /* store None as NULL */
 	else {
-	    type = pyg_type_from_object((PyObject *) item->ob_type);
+	    type = pyg_type_from_object((PyObject*)Py_TYPE(item));
 	    if (! type) {
 		PyErr_Clear();
 		g_value_array_free(value_array);
@@ -683,21 +657,25 @@ pyg_value_from_pyobject(GValue *value, PyObject *obj)
     case G_TYPE_INTERFACE:
 	/* we only handle interface types that have a GObject prereq */
 	if (g_type_is_a(G_VALUE_TYPE(value), G_TYPE_OBJECT)) {
-	    if (!PyObject_TypeCheck(obj, &PyGObject_Type)) {
-		return -1;
+	    if (obj == Py_None)
+		g_value_set_object(value, NULL);
+	    else {
+		if (!PyObject_TypeCheck(obj, &PyGObject_Type)) {
+		    return -1;
+		}
+		if (!G_TYPE_CHECK_INSTANCE_TYPE(pygobject_get(obj),
+						G_VALUE_TYPE(value))) {
+		    return -1;
+		}
+		g_value_set_object(value, pygobject_get(obj));
 	    }
-	    if (!G_TYPE_CHECK_INSTANCE_TYPE(pygobject_get(obj),
-					    G_VALUE_TYPE(value))) {
-		return -1;
-	    }
-	    g_value_set_object(value, pygobject_get(obj));
 	} else {
 	    return -1;
 	}
 	break;
     case G_TYPE_CHAR:
 	if ((tmp = PyObject_Str(obj)))
-	    g_value_set_char(value, PyString_AsString(tmp)[0]);
+	    g_value_set_char(value, _PyUnicode_AsString(tmp)[0]);
 	else {
 	    PyErr_Clear();
 	    return -1;
@@ -706,7 +684,7 @@ pyg_value_from_pyobject(GValue *value, PyObject *obj)
 	break;
     case G_TYPE_UCHAR:
 	if ((tmp = PyObject_Str(obj)))
-	    g_value_set_char(value, PyString_AsString(tmp)[0]);
+	    g_value_set_char(value, _PyUnicode_AsString(tmp)[0]);
 	else {
 	    PyErr_Clear();
 	    return -1;
@@ -717,7 +695,7 @@ pyg_value_from_pyobject(GValue *value, PyObject *obj)
 	g_value_set_boolean(value, PyObject_IsTrue(obj));
 	break;
     case G_TYPE_INT:
-	g_value_set_int(value, PyInt_AsLong(obj));
+	g_value_set_int(value, _PyLong_AsLong(obj));
 	if (PyErr_Occurred()) {
 	    g_value_unset(value);
 	    PyErr_Clear();
@@ -726,10 +704,10 @@ pyg_value_from_pyobject(GValue *value, PyObject *obj)
 	break;
     case G_TYPE_UINT:
 	{
-	    if (PyInt_Check(obj)) {
+	    if (_PyLong_Check(obj)) {
 		glong val;
 
-		val = PyInt_AsLong(obj);
+		val = _PyLong_AsLong(obj);
 		if (val >= 0 && val <= G_MAXUINT)
 		    g_value_set_uint(value, (guint)val);
 		else
@@ -745,7 +723,7 @@ pyg_value_from_pyobject(GValue *value, PyObject *obj)
 	}
 	break;
     case G_TYPE_LONG:
-	g_value_set_long(value, PyInt_AsLong(obj));
+	g_value_set_long(value, _PyLong_AsLong(obj));
 	if (PyErr_Occurred()) {
 	    g_value_unset(value);
 	    PyErr_Clear();
@@ -754,10 +732,10 @@ pyg_value_from_pyobject(GValue *value, PyObject *obj)
 	break;
     case G_TYPE_ULONG:
 	{
-	    if (PyInt_Check(obj)) {
+	    if (_PyLong_Check(obj)) {
 		glong val;
 
-		val = PyInt_AsLong(obj);
+		val = _PyLong_AsLong(obj);
 		if (val >= 0)
 		    g_value_set_ulong(value, (gulong)val);
 		else
@@ -781,8 +759,8 @@ pyg_value_from_pyobject(GValue *value, PyObject *obj)
 	}
 	break;
     case G_TYPE_UINT64:
-        if (PyInt_Check(obj))
-            g_value_set_uint64(value, PyInt_AsLong(obj));
+        if (_PyLong_Check(obj))
+            g_value_set_uint64(value, _PyLong_AsLong(obj));
         else if (PyLong_Check(obj))
             g_value_set_uint64(value, PyLong_AsUnsignedLongLong(obj));
         else
@@ -828,7 +806,7 @@ pyg_value_from_pyobject(GValue *value, PyObject *obj)
 	if (obj == Py_None)
 	    g_value_set_string(value, NULL);
 	else if ((tmp = PyObject_Str(obj))) {
-	    g_value_set_string(value, PyString_AsString(tmp));
+	    g_value_set_string(value, _PyUnicode_AsString(tmp));
 	    Py_DECREF(tmp);
 	} else {
 	    PyErr_Clear();
@@ -860,7 +838,7 @@ pyg_value_from_pyobject(GValue *value, PyObject *obj)
             GType type;
             GValue *n_value; 
             
-            type = pyg_type_from_object((PyObject *)obj->ob_type);
+            type = pyg_type_from_object((PyObject*)Py_TYPE(obj));
             if (G_UNLIKELY (! type)) {
                 PyErr_Clear();
                 return -1;
@@ -873,12 +851,12 @@ pyg_value_from_pyobject(GValue *value, PyObject *obj)
         else if (PySequence_Check(obj) &&
 		   G_VALUE_HOLDS(value, G_TYPE_VALUE_ARRAY))
 	    return pyg_value_array_from_pyobject(value, obj, NULL);
-	else if (PyString_Check(obj) &&
+	else if (_PyUnicode_Check(obj) &&
                  G_VALUE_HOLDS(value, G_TYPE_GSTRING)) {
             GString *string;
             char *buffer;
             Py_ssize_t len;
-            if (PyString_AsStringAndSize(obj, &buffer, &len))
+            if (_PyUnicode_AsStringAndSize(obj, &buffer, &len))
                 return -1;
             string = g_string_new_len(buffer, len);
 	    g_value_set_boxed(value, string);
@@ -942,17 +920,17 @@ pyg_value_as_pyobject(const GValue *value, gboolean copy_boxed)
 	    break;
     case G_TYPE_CHAR: {
 	gint8 val = g_value_get_char(value);
-	return PyString_FromStringAndSize((char *)&val, 1);
+	return _PyUnicode_FromStringAndSize((char *)&val, 1);
     }
     case G_TYPE_UCHAR: {
 	guint8 val = g_value_get_uchar(value);
-	return PyString_FromStringAndSize((char *)&val, 1);
+	return _PyUnicode_FromStringAndSize((char *)&val, 1);
     }
     case G_TYPE_BOOLEAN: {
 	return PyBool_FromLong(g_value_get_boolean(value));
     }
     case G_TYPE_INT:
-	return PyInt_FromLong(g_value_get_int(value));
+	return _PyLong_FromLong(g_value_get_int(value));
     case G_TYPE_UINT:
 	{
 	    /* in Python, the Int object is backed by a long.  If a
@@ -960,19 +938,19 @@ pyg_value_as_pyobject(const GValue *value, gboolean copy_boxed)
 	       an Int.  Otherwise, use a Long object to avoid overflow.
 	       This matches the ULongArg behavior in codegen/argtypes.h */
 #if (G_MAXUINT <= G_MAXLONG)
-	    return PyInt_FromLong((glong) g_value_get_uint(value));
+	    return _PyLong_FromLong((glong) g_value_get_uint(value));
 #else
 	    return PyLong_FromUnsignedLong((gulong) g_value_get_uint(value));
 #endif
 	}
     case G_TYPE_LONG:
-	return PyInt_FromLong(g_value_get_long(value));
+	return _PyLong_FromLong(g_value_get_long(value));
     case G_TYPE_ULONG:
 	{
 	    gulong val = g_value_get_ulong(value);
 
 	    if (val <= G_MAXLONG)
-		return PyInt_FromLong((glong) val);
+		return _PyLong_FromLong((glong) val);
 	    else
 		return PyLong_FromUnsignedLong(val);
 	}
@@ -981,7 +959,7 @@ pyg_value_as_pyobject(const GValue *value, gboolean copy_boxed)
 	    gint64 val = g_value_get_int64(value);
 
 	    if (G_MINLONG <= val && val <= G_MAXLONG)
-		return PyInt_FromLong((glong) val);
+		return _PyLong_FromLong((glong) val);
 	    else
 		return PyLong_FromLongLong(val);
 	}
@@ -990,7 +968,7 @@ pyg_value_as_pyobject(const GValue *value, gboolean copy_boxed)
 	    guint64 val = g_value_get_uint64(value);
 
 	    if (val <= G_MAXLONG)
-		return PyInt_FromLong((glong) val);
+		return _PyLong_FromLong((glong) val);
 	    else
 		return PyLong_FromUnsignedLongLong(val);
 	}
@@ -1007,7 +985,7 @@ pyg_value_as_pyobject(const GValue *value, gboolean copy_boxed)
 	    const gchar *str = g_value_get_string(value);
 
 	    if (str)
-		return PyString_FromString(str);
+		return _PyUnicode_FromString(str);
 	    Py_INCREF(Py_None);
 	    return Py_None;
 	}
@@ -1037,7 +1015,7 @@ pyg_value_as_pyobject(const GValue *value, gboolean copy_boxed)
 	    return ret;
 	} else if (G_VALUE_HOLDS(value, G_TYPE_GSTRING)) {
 	    GString *string = (GString *) g_value_get_boxed(value);
-	    PyObject *ret = PyString_FromStringAndSize(string->str, string->len);
+	    PyObject *ret = _PyUnicode_FromStringAndSize(string->str, string->len);
 	    return ret;
 	}	    
 	bm = pyg_type_lookup(G_VALUE_TYPE(value));
@@ -1511,48 +1489,12 @@ object_doc_descr_get(PyObject *self, PyObject *obj, PyObject *type)
         g_array_free(parents, TRUE);
     }
 
-    pystring = PyString_FromStringAndSize(string->str, string->len);
+    pystring = _PyUnicode_FromStringAndSize(string->str, string->len);
     g_string_free(string, TRUE);
     return pystring;
 }
 
-static PyTypeObject PyGObjectDoc_Type = {
-    PyObject_HEAD_INIT(NULL)
-    0,
-    "gobject.GObject__doc__",
-    sizeof(PyObject),
-    0,
-    (destructor)object_doc_dealloc,
-    (printfunc)0,
-    (getattrfunc)0,
-    (setattrfunc)0,
-    (cmpfunc)0,
-    (reprfunc)0,
-    0,
-    0,
-    0,
-    (hashfunc)0,
-    (ternaryfunc)0,
-    (reprfunc)0,
-    (getattrofunc)0,
-    (setattrofunc)0,
-    0,
-    Py_TPFLAGS_DEFAULT,
-    NULL,
-    (traverseproc)0,
-    (inquiry)0,
-    (richcmpfunc)0,
-    0,
-    (getiterfunc)0,
-    (iternextfunc)0,
-    0,
-    0,
-    0,
-    (PyTypeObject *)0,
-    (PyObject *)0,
-    (descrgetfunc)object_doc_descr_get,
-    (descrsetfunc)0
-};
+PYGLIB_DEFINE_TYPE("gobject.GObject.__doc__", PyGObjectDoc_Type, PyObject);
 
 /**
  * pyg_object_descr_doc_get:
@@ -1569,7 +1511,7 @@ pyg_object_descr_doc_get(void)
     static PyObject *doc_descr = NULL;
 
     if (!doc_descr) {
-	PyGObjectDoc_Type.ob_type = &PyType_Type;
+	Py_TYPE(&PyGObjectDoc_Type) = &PyType_Type;
 	if (PyType_Ready(&PyGObjectDoc_Type))
 	    return NULL;
 
@@ -1732,3 +1674,63 @@ _pyg_type_from_name(const gchar *name)
     return type;
 }
 
+static PyObject *
+_pyg_strv_from_gvalue(const GValue *value)
+{
+    gchar    **argv = (gchar **) g_value_get_boxed(value);
+    int        argc = 0, i;
+    PyObject  *py_argv;
+
+    if (argv) {
+        while (argv[argc])
+            argc++;
+    }
+    py_argv = PyList_New(argc);
+    for (i = 0; i < argc; ++i)
+	PyList_SET_ITEM(py_argv, i, _PyUnicode_FromString(argv[i]));
+    return py_argv;
+}
+
+static int
+_pyg_strv_to_gvalue(GValue *value, PyObject *obj)
+{
+    Py_ssize_t argc, i;
+    gchar **argv;
+
+    if (!(PyTuple_Check(obj) || PyList_Check(obj)))
+        return -1;
+
+    argc = PySequence_Length(obj);
+    for (i = 0; i < argc; ++i)
+	if (!_PyUnicode_Check(PySequence_Fast_GET_ITEM(obj, i)))
+	    return -1;
+    argv = g_new(gchar *, argc + 1);
+    for (i = 0; i < argc; ++i)
+	argv[i] = g_strdup(_PyUnicode_AsString(PySequence_Fast_GET_ITEM(obj, i)));
+    argv[i] = NULL;
+    g_value_take_boxed(value, argv);
+    return 0;
+}
+
+void
+pygobject_type_register_types(PyObject *d)
+{
+    PyGTypeWrapper_Type.tp_dealloc = (destructor)pyg_type_wrapper_dealloc;
+    PyGTypeWrapper_Type.tp_compare = (cmpfunc)pyg_type_wrapper_compare;
+    PyGTypeWrapper_Type.tp_repr = (reprfunc)pyg_type_wrapper_repr;
+    PyGTypeWrapper_Type.tp_hash = (hashfunc)pyg_type_wrapper_hash;
+    PyGTypeWrapper_Type.tp_flags = Py_TPFLAGS_DEFAULT;
+    PyGTypeWrapper_Type.tp_methods = _PyGTypeWrapper_methods;
+    PyGTypeWrapper_Type.tp_getset = _PyGTypeWrapper_getsets;
+    PyGTypeWrapper_Type.tp_init = (initproc)pyg_type_wrapper_init;
+    PYGLIB_REGISTER_TYPE(d, PyGTypeWrapper_Type, "GType");
+
+    /* This type lazily registered in pyg_object_descr_doc_get */
+    PyGObjectDoc_Type.tp_dealloc = (destructor)object_doc_dealloc;
+    PyGObjectDoc_Type.tp_flags = Py_TPFLAGS_DEFAULT;
+    PyGObjectDoc_Type.tp_descr_get = (descrgetfunc)object_doc_descr_get;
+    
+    pyg_register_gtype_custom(G_TYPE_STRV,
+			      _pyg_strv_from_gvalue,
+			      _pyg_strv_to_gvalue);
+}
